from CPR import tools
from CPR import taxes
from CPR import simulator


def compute_bs_bef_ret(hh, year, common, prices):
    """
    Function to compute the pre-retirement balance sheet.

    Parameters
    ----------
    hh: Hhold
        household
    year : int
        year
    common : Common
        instance of the class Common
    prices : Prices
        instance of the class Prices
    """
    nom = tools.create_nom(year, prices)

    hh_srd = taxes.file_household(hh, year, common, prices)

    for who, p in enumerate(hh.sp):
        p.disp_inc_bef_ret = nom(hh_srd.sp[who].disp_inc)
    compute_cons_bef_ret(hh, year, prices)


def compute_cons_bef_ret(hh, year, prices):
    """
    Function to compute pre-retirement consumption.

    Parameters
    ----------
    hh: Hhold
        household
    year : int
        year
    prices : Prices
        instance of the class Prices
    """
    real = tools.create_real(year, prices)

    hh.disp_inc_bef_ret = sum([p.disp_inc_bef_ret for p in hh.sp])
    hh.debt_payments = sum([hh.debts[debt].payment for debt in hh.debts])
    hh.cons_bef_ret_real = real(hh.disp_inc_bef_ret - hh.debt_payments)


def compute_bs_after_ret(hh, year, common, prices):
    """
    Function to compute the post-retirement balance sheet.

    Parameters
    ----------
    hh: Hhold
        household
    year : int
        year
    common : Common
        instance of the class Common
    prices : Prices
        instance of the class Prices
    """
    nom, real = tools.create_nom_real(year, prices)

    hh_tax = taxes.file_household(hh, year, common, prices)
    for inc in ['fam_inc_tot', 'fam_after_tax_inc', 'fam_disp_inc']:
        val = nom(getattr(hh_tax, inc))
        setattr(hh, f'{inc}_after_ret', val)
    taxes.get_gis_oas_allowances(hh, hh_tax, year, prices)

    hh.debt_payments = sum([hh.debts[debt].payment for debt in hh.debts])
    hh.cons_after_ret_real = real(hh.fam_disp_inc_after_ret - hh.debt_payments)
    hh.cons_after_ret_real -= real(getattr(hh, 'imputed_rent', 0))


def add_output(hh, year, prices, key):
    """
    Function to extract output variables.

    Parameters
    ----------
    hh: Hhold
        household
    year : int
        year
    prices : Prices
        instance of the class Prices
    key : str
        before ("bef"), when first spouse retires ("part") 
        or after retirement ("aft")
    """
    real = tools.create_real(year, prices)

    for p in hh.sp:
        hh.d_output[f'{p.who}wage_{key}'] = real(p.d_wages[year])
        hh.d_output[f'{p.who}pension_{key}'] = real(p.pension)

    for residence in hh.residences:
        hh.d_output[f'{residence}_{key}'] = \
            real(hh.residences[residence].balance)    
    business = real(hh.business.balance) if hasattr(hh, 'business') else 0
    hh.d_output[f'business_{key}'] = business
        
    if 'first_mortgage' in hh.debts:
        hh.d_output[f'first_mortgage_balance_{key}'] = \
            real(hh.debts['first_mortgage'].balance)

    if key == 'bef':
        hh.d_output[f'year_cons_bef'] = hh.cons_bef_ret_year
        hh.d_output[f'cons_{key}'] = hh.cons_bef_ret_real

    if key in ['bef', 'part']:
        for p in hh.sp:
            rpp_dc = real(p.rpp_dc.balance) if hasattr(p, 'rpp_dc') else 0
            hh.d_output[f'{p.who}rpp_dc_{key}'] = rpp_dc
            for acc in p.fin_assets:
                hh.d_output[f'{p.who}{acc}_balance_{key}'] = \
                    real(p.fin_assets[acc].balance)

    if key in ['part', 'after']:
        for p in hh.sp:
            hh.d_output[f'{p.who}annuity_rrsp_{key}'] = p.annuity_rrsp_real
            hh.d_output[f'{p.who}annuity_rpp_dc_{key}'] = p.annuity_rpp_dc_real
            hh.d_output[f'{p.who}annuity_non_rrsp_{key}'] = \
                p.annuity_non_rrsp_real

    if key == 'after':
        hh.d_output[f'year_cons_after'] = hh.cons_after_ret_year
        hh.d_output[f'imputed_rent_{key}'] = real(getattr(hh, 'imputed_rent', 0))
        hh.d_output[f'cons_{key}'] = hh.cons_after_ret_real
        hh.d_output[f'debt_payments_{key}'] = real(hh.debt_payments)
        hh.d_output[f'fam_net_tax_liability_{key}'] = real(
            hh.fam_inc_tot_after_ret - hh.fam_after_tax_inc_after_ret)

        for p in hh.sp:
            hh.d_output[f'{p.who}cpp_{key}'] = real(p.cpp)
            hh.d_output[f'{p.who}gis_{key}'] = real(p.inc_gis)
            hh.d_output[f'{p.who}oas_{key}'] = real(p.inc_oas)
            hh.d_output[f'{p.who}allow_surv_{key}'] = real(p.allow_surv)
            hh.d_output[f'{p.who}allow_couple_{key}'] = real(p.allow_couple)          
            db_benefits = real(p.rpp_db.benefits) if hasattr(p, 'rpp_db') else 0
            hh.d_output[f'{p.who}rpp_db_benefits_{key}'] = db_benefits
            hh.d_output[f'{p.who}business_dividends_{key}'] = real(getattr(p, 'div_other_can', 0))


import numpy as np
import srpp
from CPR import annuities
from CPR import taxes
from CPR import tools
from CPR import balance_sheets


def simulate(job, common, prices):
    """
    Function that projects assets, RPPs and debts until the time of retirement.

    Parameters
    ----------
    job: tuple(Hhold, int)
        instance of the class Hhold and simulation number
    common: Common
        instance of the class Common
    prices: Prices
        instance of the class Prices

    Returns
    -------
    dict:
        dictionary containing households' characteristics
        before and after retirement
    """
    hh, sim = job

    # store output in dictionary:
    hh.d_output = dict(hh_index=hh.index, sim=sim)

    d_returns, prices.d_price_rent_ratio = extract_time_series(sim, common,
                                                               prices)

    for p in hh.sp:
        prepare_wages(p, sim, common, prices)
        initialize_cpp_account(p, hh, common)

    year = common.base_year
    while year < hh.ret_year:
        update_ages(hh, year)
        update_debts(hh, year, sim, common, prices)
        adjust_contributions(hh, year, common, prices)
        update_assets(hh, year, d_returns, common, prices)
        manage_liquidations(hh, year, common, prices)

        for p in hh.sp:
            contribute_cpp(p, year, common)
            claim_cpp(p)

        if check_liquidation(hh) or check_tax_unreg(hh):
            prepare_taxes(hh, year, common, prices)
            taxes.compute_after_tax_amount(hh, year, common, prices)

        if year == hh.cons_bef_ret_year:
            prepare_taxes(hh, year, common, prices)
            balance_sheets.compute_bs_bef_ret(hh, year, common, prices)
            balance_sheets.add_output(hh, year, prices, 'bef')

        # partial retirement (some couples)
        if year == getattr(hh, 'partial_ret_year', np.nan) - 1:
            annuities.compute_partial_annuities(hh, d_returns, year, prices)
        if year == getattr(hh, 'partial_ret_year', np.nan):
            balance_sheets.add_output(hh, year, prices, 'part')

        # full retirement
        if year == hh.ret_year - 1:
            annuities.compute_annuities(hh, d_returns, year, prices)

        year += 1

    while True:
        update_ages(hh, year)
        update_debts(hh, year, sim, common, prices)
        for p in hh.sp:
            claim_cpp(p)
        if year == hh.cons_after_ret_year:
            break
        year += 1

    prepare_taxes(hh, year, common, prices)
    balance_sheets.compute_bs_after_ret(hh, year, common, prices)
    balance_sheets.add_output(hh, year, prices, 'after')

    # reset assets, debts and annuities
    reset_accounts(hh)

    return hh.d_output


def extract_time_series(sim, common, prices):
    """
    Function that attaches stochastic processes on asset returns to households.

    Parameters
    ----------
    sim : int
        simulation number
    common : Common
        instance of the class Common
    prices : Prices
        instance of the class Prices

    Returns
    -------
    dict
        returns on assets
    dict
        price-rent ratio
    """
    d_returns = {}
    for ret in ['bills', 'bonds', 'equity', 'housing', 'business']:
        years = range(common.base_year, common.base_year + common.future_years)
        d_returns[ret] = dict(zip(years,
                                  getattr(prices, f'ret_{ret}')[:, sim]))
    d_price_rent_ratio = dict(zip(years, prices.price_rent_ratio[:, sim]))

    return d_returns, d_price_rent_ratio


def prepare_wages(p, sim, common, prices):
    """
    Function that attaches wage profiles to individuals.

    Parameters
    ----------
    p: Person
        instance of the class Person
    sim : int
        simulation number
    common : Common
        instance of the class Common
    prices: Prices
        instance of the class Prices
    """
    time_range = range(p.byear + common.min_age_cpp,
                       p.byear + common.future_years)
    p.d_wages = dict(zip(time_range, p.wage_profile[:, sim]))
    p.d_wages = {year: p.d_wages[year] * prices.d_infl_factors[year]
                 for year in p.d_wages}


def initialize_cpp_account(p, hh, common):
    """
    Function that creates a CPP/QPP account and enters past CPP/QPP contributions based on past wages.

    Parameters
    ----------
    p: Person
        instance of the class Person
    hh: Hhold
        household
    common : Common
        instance of the class Common
    """

    rules = common.rules_qpp if hh.prov == 'qc' else common.rules_cpp
    p.cpp_account = srpp.account(p.byear, rules)

    for year in range(p.byear + common.min_age_cpp, p.byear + p.age):
        p.cpp_account.MakeContrib(year, p.d_wages[year])


def update_ages(hh, year):
    """
    Function that computes age for a given year.

    Parameters
    ----------
    hh: Hhold
        household
    year : int
        year
    """
    for p in hh.sp:
        p.age = year - p.byear


def update_debts(hh, year, sim, common, prices):
    """
    Function that updates debt payments and balances.

    Parameters
    ----------
    hh: Hhold
        household
    year : int
        year
    sim : int
        simulation number
    common : Common
        instance of the class Common
    prices: Prices
        instance of the class Prices
    """
    for debt in hh.debts:
        hh.debts[debt].update(
            year, prices.d_interest_debt[debt][year - common.base_year, sim],
            prices)


def adjust_contributions(hh, year, common, prices):
    """
    Function that adjusts contributions to all account types (RRSP, other registered, TFSA) and to pensions plans (DC and DB RPPs) so as to respect available contribution rooms.

    Parameters
    ----------
    hh: Hhold
        household
    year : int
        year
    common : Common
        instance of the class Common
    prices: Prices
        instance of the class Prices
    """
    for p in hh.sp:
        if (p.replacement_rate_db > 0):
            if p.age >= (p.init_age + int(p.replacement_rate_db
                                          / common.perc_year_db)):
                p.rpp_db.rate_employee_db = 0
        if not p.retired:
            p.contribution_room.compute_contributions(p, year, common, prices)


def update_assets(hh, year, d_returns, common, prices):
    """
    Function that updates assets every year.

    Parameters
    ----------
    hh: Hhold
        household
    year : int
        year
    d_returns : dict
        dictionary of returns
    common : Common
        instance of the class Common
    prices: Prices
        instance of the class Prices
    """
    for p in hh.sp:
        if p.retired:
            continue
        if hasattr(p, 'rpp_dc'):
            p.rpp_dc.update(d_returns, year, common, prices)
        for acc in p.fin_assets:
            p.fin_assets[acc].update(d_returns, year, common, prices)
    for acc in hh.residences:
        hh.residences[acc].update(d_returns['housing'], year, prices)
    if hasattr(hh, 'business'):
        hh.business.update(d_returns['business'], year, prices)


def manage_liquidations(hh, year, common, prices):
    """
    Function that manages the liquidation of assets upon retirement.

    Parameters
    ----------
    hh: Hhold
        household
    year : int
        year
    common : Common
        instance of the class Common
    prices: Prices
        instance of the class Prices
    """
    nom = tools.create_nom(year, prices)

    for p in hh.sp:
        p.liquidation_non_taxable = 0
        p.liquidation_cap_gains = 0
        p.liquidation_cap_losses = 0
        p.liquidation_business_exempt = 0

    if year == getattr(hh, 'partial_ret_year', np.nan) - 1:
        unreg = hh.sp[0].fin_assets['unreg']
        unreg.liquidate()
        hh.sp[0].liquidation_non_taxable += unreg.liquidation_non_taxable
        hh.sp[0].liquidation_cap_gains += unreg.liquidation_cap_gains
        hh.sp[0].liquidation_cap_losses += unreg.liquidation_cap_losses

    if year == hh.ret_year - 1:
        for p in hh.sp:
            unreg = p.fin_assets['unreg']
            unreg.liquidate()
            p.liquidation_non_taxable += unreg.liquidation_non_taxable
            p.liquidation_cap_gains += unreg.liquidation_cap_gains
            p.liquidation_cap_losses += unreg.liquidation_cap_losses

        if common.sell_first_resid & ('first_residence' in hh.residences):
            first_res = hh.residences['first_residence']
            first_res.impute_rent(hh, year, common, prices)
            first_res.liquidate()
            liquidation_value = (first_res.liquidation_non_taxable
                                 + first_res.liquidation_cap_gains)
            if 'first_mortgage' in hh.debts:
                liquidation_value -= hh.debts['first_mortgage'].balance
                hh.debts['first_mortgage'].balance = 0
            for p in hh.sp:
                p.liquidation_non_taxable += (liquidation_value
                                              / (1 + hh.couple))

        if common.sell_second_resid & ('second_residence' in hh.residences):
            second_res = hh.residences['second_residence']
            second_res.liquidate()
            if 'second_mortgage' in hh.debts:
                second_res.liquidation_non_taxable -= (
                    hh.debts['second_mortgage'].balance)
                hh.debts['second_mortgage'].balance = 0
            for p in hh.sp:
                p.liquidation_non_taxable += (
                    second_res.liquidation_non_taxable / (1 + hh.couple))
                p.liquidation_cap_gains += (
                    second_res.liquidation_cap_gains / (1 + hh.couple))

        if common.sell_business & hasattr(hh, 'business'):
            hh.business.liquidate(common)
            liquidation_business_exempt = min(
                hh.business.liquidation_cap_gains, nom(common.lcge_real))
            for p in hh.sp:
                p.liquidation_non_taxable += (
                    hh.business.liquidation_non_taxable / (1 + hh.couple))
                p.liquidation_cap_gains += (
                    hh.business.liquidation_cap_gains / (1 + hh.couple))
                p.liquidation_business_exempt += (
                        liquidation_business_exempt / (1 + hh.couple))


def contribute_cpp(p, year, common):
    """
    Function that records CPP/QPP contributions.

    Parameters
    ----------
    p: Person
        instance of the class Person
    year : int
        year
    common : Common
        instance of the class Common
    """
    if (not p.retired) & (p.age < common.max_claim_age_cpp):
        p.cpp_account.MakeContrib(year, p.d_wages[year])


def claim_cpp(p):
    """
    Function to claim CPP/QPP benefits.

    Parameters
    ----------
    p: Person
        instance of the class Person
    """
    if p.age == p.claim_age_cpp:
        p.cpp_account.ClaimCPP(p.byear + p.claim_age_cpp)


def check_tax_unreg(hh):
    """
    Function that checks whether there are taxable returns to assets.

    Parameters
    ----------
    hh: Hhold
        household

    Returns
    -------
    bool
        True or False
    """
    for p in hh.sp:
        if p.fin_assets['unreg'].income_to_tax:
            return True
    return False


def check_liquidation(hh):
    """
    Function to check whether there are liquidated assets to be taxed.

    Parameters
    ----------
    hh: Hhold
        household

    Returns
    -------
    bool
        True or False
    """
    for p in hh.sp:
        p.liquidation_to_tax = (p.liquidation_non_taxable > 0 or
                                p.liquidation_cap_gains > 0 or
                                p.liquidation_cap_losses > 0 or
                                p.liquidation_business_exempt > 0)
    for p in hh.sp:
        if p.liquidation_to_tax:
            return True
    return False


def prepare_taxes(hh, year, common, prices):
    """
    Function to prepare the variables used in the SRD (in nominal terms).

    Parameters
    ----------
    hh: Hhold
        household
    year : int
        year
    common : Common
        instance of the class Common
    prices : Prices
        instance of the class Prices
    """
    nom = tools.create_nom(year, prices)

    for p in hh.sp:
        p.earn = p.d_wages[year]
        compute_rpp(p, nom, common)
        get_benefits_cpp(p, year, common)
        p.net_cap_gains = p.fin_assets['unreg'].withdrawal_cap_gains
        p.prev_cap_losses = p.fin_assets['unreg'].withdrawal_cap_losses
        p.div_elig = (common.share_div_elig
                      * p.fin_assets['unreg'].withdrawal_div)
        p.div_other_can = (hh.business.dividends_business / (1 + hh.couple)
                           if hasattr(hh, 'business') else 0)
        get_other_taxable(p, nom, common)
        get_other_non_taxable(p, nom)
        get_inc_rrsp(p, nom)
        get_contributions_assets(p, year, common)
        get_assets(p, common)


def get_assets(p, common):
    """
    Function to retrieve assets for social assistance asset test.

    Parameters
    ----------
    p: Person
        instance of the class Person
    common : Common
        instance of class Common
    """
    assets_unreg, assets_reg = 0, 0
    for acc in p.fin_assets:
        if acc == 'unreg':
            assets_unreg += p.fin_assets['unreg'].balance
        else:
            assets_reg += p.fin_assets[acc].balance
    p.assets = assets_unreg + max(0, assets_reg - common.exempt_soc_ass)

def compute_rpp(p, nom, common):
    """
    Compute rpp (DB and pension).

    Parameters
    ----------
    p: Person
        instance of the class Person
    nom: function
        function converting to nominal value
    common: Common
        instance of the class Common
    """
    p.rpp = nom(p.annuity_rpp_dc_real)
    if p.pension > 0:
        p.rpp += p.pension
    if hasattr(p, 'rpp_db') & p.retired & (p.age > common.db_minimum_age):
        if p.rpp_db.benefits == 0: # avoid multiple counting if spouses retire in different years (and have unregistered assets)
            p.rpp_db.compute_benefits(p, common)
        p.rpp += p.rpp_db.benefits


def get_benefits_cpp(p, year, common):
    """
    Function that computes annual CPP/QPP retirement benefits.

    s1 is the part of the supplementary benefit due to higher contribution rates in the 2019-2025 expansion; s2 is the part of the supplementary benefits attributable to changes in the YMPE; and PRB is for post-retirement benefits (for individuals who keep working after claiming -- this feature is not used in the CPR, since individuals automatically claim in the year they retire).

    Parameters
    ----------
    p: Person
        instance of the class Person
    year : int
        year
    common : Common
        instance of the class Common
    """
    p.cpp = 12 * (
        p.cpp_account.gBenefit(year)
        + p.cpp_account.gBenefit_s1(year) + p.cpp_account.gBenefit_s2(year)
        + p.cpp_account.gPRB(year)
        + p.cpp_account.gPRB_s1(year) + p.cpp_account.gPRB_s2(year))

    if common.old_cpp:
        p.cpp -= 12 * (
            p.cpp_account.gBenefit_s1(year) + p.cpp_account.gBenefit_s2(year)
            + p.cpp_account.gPRB_s1(year) + p.cpp_account.gPRB_s2(year))


def get_inc_rrsp(p, nom):
    """
    Function that computes RRSP income from withdrawals.

    Parameters
    ----------
    p: Person
        instance of the class Person
    nom: function
        function converting to nominal value
    """
    p.inc_rrsp = nom(p.annuity_rrsp_real)

    if p.retired:
        return

    for acc in set(p.fin_assets) & set(('rrsp', 'other_reg')):
        p.inc_rrsp += p.fin_assets[acc].withdrawal


def get_other_taxable(p, nom, common):
    """
    Function that computes other taxable income.

    Parameters
    ----------
    p: Person
        instance of the class Person
    nom: function
        function converting to nominal value
    common : Common
        instance of the class Common
    """
    regular_div_and_int = (
        (1 - common.share_div_elig) * p.fin_assets['unreg'].withdrawal_div
        + p.fin_assets['unreg'].withdrawal_int)
    annuity_return = nom(p.annuity_non_rrsp_real - p.annuity_non_rrsp_0_real)
    p.other_taxable = regular_div_and_int + annuity_return


def get_other_non_taxable(p, nom):
    """
    Function that computes other non-taxable income.

    Parameters
    ----------
    p: Person
        instance of the class Person
    nom: function
        function converting to nominal value
    """
    withdrawal_non_tax = (p.fin_assets['unreg'].withdrawal_non_tax
                          + p.fin_assets['tfsa'].withdrawal)
    annuity_non_tax = nom(p.annuity_non_rrsp_0_real)

    p.other_non_taxable = withdrawal_non_tax + annuity_non_tax


def get_contributions_assets(p, year, common):
    """
    Function that computes contributions to registered and unregistered accounts (including DC and DB RPP).

    Parameters
    ----------
    p: Person
        instance of the class Person
    year : int
        year
    common : Common
        instance of the class Common
    """
    p.con_rrsp, p.con_non_rrsp = 0, 0

    if p.retired:
        return

    for acc in set(p.fin_assets):
        if acc in ['rrsp', 'other_reg']:
            p.con_rrsp += p.fin_assets[acc].contribution
        else:
            p.con_non_rrsp += p.fin_assets[acc].contribution
    if (p.rate_employee_db > 0) & (year < common.base_year
                                   + common.max_years_db):
        p.con_rrsp += p.rate_employee_db * p.d_wages[year]
    if hasattr(p, 'rpp_dc'):
        p.con_rrsp += p.contrib_employee_dc


def reset_accounts(hh):
    """
    Function that resets all variables to their initial values.

    Parameters
    ----------
    hh: Hhold
        household
    """
    for p in hh.sp:
        p.age = p.init_age
        p.retired = False
        p.contribution_room.reset()
        # assets:
        if hasattr(p, 'rpp_dc'):
            p.rpp_dc.reset()
        if hasattr(p, 'rpp_db'):
            p.rpp_db.reset()
        for acc in p.fin_assets:
            p.fin_assets[acc].reset()
    for acc in hh.residences:
        hh.residences[acc].reset()
    if hasattr(hh, 'business'):
        hh.business.reset()
    # debts
    for debt in hh.debts:
        hh.debts[debt].reset()
    # annuities
    for p in hh.sp:
        for type in ['non_rrsp_0', 'non_rrsp', 'rpp_dc', 'rrsp']:
            setattr(p, f'annuity_{type}_real', 0)
    # cpp/qpp
        p.cpp_account.ResetCase()
        p.cpp = 0



from CPR import taxes
from CPR import tools


def compute_partial_annuities(hh, d_returns, year, prices):
    """
    Function that partially converts assets into annuities when the first spouse retires, using the other functions below.

    Parameters
    ----------
    hh: Hhold
        household
    d_returns : dict
        dictionary of returns
    year : int
        year
    prices : Prices
        instance of the class Prices
    """
    liquidate_fin_assets(hh.sp[0])
    for acc in ['rrsp', 'rpp_dc', 'tfsa', 'unreg']:
        setattr(hh.sp[1], f'val_annuities_{acc}', 0)
    compute_factors(hh, hh.sp[0], d_returns['bonds'][year], prices)
    convert_to_real_annuities(hh.sp[0], year, prices)
    hh.sp[0].retired = True


def compute_annuities(hh, d_returns, year, prices):
    """
    Function that fully converts assets into annuities when the last spouse retires, using the other functions below.

    Parameters
    ----------
    hh: Hhold
        household
    d_returns : dict
        dictionary of returns
    year : int
        year
    prices : Prices
        instance of the class Prices
    """
    for p in hh.sp:
        liquidate_fin_assets(p)
        compute_factors(hh, p, d_returns['bonds'][year], prices)
        convert_to_real_annuities(p, year, prices)
        p.retired = True


def liquidate_fin_assets(p):
    """
    Function to liquidate financial assets.

    Parameters
    ----------
    p: Person
        instance of the class Person
    """
    for acc in ['rrsp', 'rpp_dc', 'tfsa']:
        setattr(p, f'val_annuities_{acc}', 0)
    for acc in p.fin_assets:
        if acc in ['rrsp', 'other_reg']:
            p.val_annuities_rrsp += p.fin_assets[acc].liquidate()
        if acc == 'tfsa':
            p.val_annuities_tfsa += p.fin_assets[acc].liquidate()
    if hasattr(p, 'rpp_dc'):
        p.val_annuities_rpp_dc += p.rpp_dc.liquidate()


def compute_factors(hh, p, rate, prices):
    """
    Function to compute an individual factor for annuities constant in real terms.

    We use an adjusted rate to dampen excess factor volatility. This is because we take the total return on bonds in the year that the annuity is purchased. Ideally, the whole interest structure should be used instead, such that mean reversion would smooth out annuity prices.

    Parameters
    ----------
    hh: Hhold
        household
    p: Person
        instance of the class Person
    rate : float
        interest rate
    prices : Prices
        instance of the class Prices
    """
    rate_real = (1 + rate) / (1 + prices.inflation_rate) - 1
    adj_rate = (
        rate_real - prices.adj_fact_annuities * (rate_real - prices.mu_bonds))
    p.factor = prices.d_factors[p.sex][hh.prov].compute_annuity_factor(
        p.byear, agestart=p.ret_age, rate=adj_rate)
    real_rate_zero_ret = 1 / (1 + prices.inflation_rate) - 1
    p.factor_0 = prices.d_factors[p.sex][hh.prov].compute_annuity_factor(
        p.byear, agestart=p.ret_age, rate=-real_rate_zero_ret)


def convert_to_real_annuities(p, year, prices):
    """
    Function that converts assets to annuities in real terms.

    Parameters
    ----------
    p: Person
        instance of the class Person
    year : int
        year
    prices : Prices
        instance of the class Prices
    """
    real = tools.create_real(year, prices)

    if hasattr(p, 'liquidation_after_tax'):
        amount, p.liquidation_after_tax = p.liquidation_after_tax, 0
    else:
        amount = 0

    p.annuity_rrsp_real += real(p.val_annuities_rrsp) / p.factor
    p.annuity_rpp_dc_real += real(p.val_annuities_rpp_dc) / p.factor
    p.annuity_non_rrsp_real += real((p.val_annuities_tfsa + amount) / p.factor)
    p.annuity_non_rrsp_0_real += real((p.val_annuities_tfsa + amount)
                                      / p.factor_0)



import os
from functools import partial
import multiprocessing as mp
import pandas as pd
from CPR import initialisation
from CPR import macro
from CPR import simulator
from CPR import tools
from CPR import analysis

module_dir = os.path.dirname(os.path.dirname(__file__))
path = '/CPR/data/params/'

def run_simulations(inputs, nsim=1, non_stochastic=False, n_jobs=None,
                    **extra_params):
    """
    This function launches the simulations. Any parameter can be changed
    using extra_params.

    Parameters
    ----------
    inputs : _io.TextIOWrapper
        csv file
    nsim : int, optional
        number of simulations, by default 1
    non_stochastic : bool, optional
        deterministic prices and price-rent ratio, by default False

    Returns
    -------
    Results
        instance of the class Results
    """
    # check extra_params in common_params or prices:
    l_params = []
    for file_csv in ['prices.csv', 'common_params.csv', 'user_options.csv']:
        params = tools.get_params(module_dir + path + file_csv)
        l_params.extend(params.keys())
    for param in extra_params:
        assert param in l_params, f'{param} is not a parameter'

    if non_stochastic:
        nsim = 1

    common = macro.CommonParameters(nsim, non_stochastic, extra_params)
    prices = macro.Prices(common, extra_params)
    
    if type(inputs) is pd.Series:
        d_input = inputs.to_frame().to_dict()
    else:
        d_input = inputs.to_dict('index')

    hhs = [initialisation.create_hh(index, d_hh, common, prices)
           for index, d_hh in d_input.items()]
    jobs = [(hh, sim) for hh in hhs for sim in range(nsim)]

    if n_jobs is None:
            n_jobs = mp.cpu_count()
    
    if n_jobs > 1:
        with mp.Pool(processes=n_jobs) as pool:
            l_outputs = pool.map(partial(simulator.simulate,
                                 common=common, prices=prices), jobs)
    else:
        l_outputs = [simulator.simulate(job, common, prices) for job in jobs]

    output = pd.DataFrame.from_dict(l_outputs)
    results = analysis.Results(inputs, output, common, prices, extra_params)
    return results



import numpy as np


class Debt():
    """
    This class manages all debts. All amounts are nominal.
    
    Parameters
    ----------
    name: str
        name of the debt
    d_hh: dict
        dictionary containing debt initial balances and monthly payments
    common: Common
        instance of the class Common
    prices: Prices
        instance of the class Prices
    """
    def __init__(self, name, d_hh, common, prices):
        self.name = name
        self.init_balance = d_hh[name]
        self.init_m_payment = d_hh[name + '_payment']
        self.init_term = self.estimate_init_term(common, prices)
        self.reset()

    def estimate_init_term(self, common, prices):
        """
        Estimate the term of the debt.

        Parameters
        ----------
        common : Common
            instance of the class Common
        prices : Prices
            instance of the class Prices

        Returns
        -------
        int
            term of the debt
        """
        init_rate = prices.d_interest_debt[self.name][0, 0]
        init_monthly_rate = (1 + init_rate)**(1/12) - 1
        interest = init_monthly_rate / (1 + init_monthly_rate) \
            * self.init_balance
        init_m_payment = max(self.init_m_payment, interest + 1e-12)
        # debt is explosive if init_payment lower than interest
        term_months = (np.log(init_m_payment / (init_m_payment - interest))
                       / np.log(1 + init_monthly_rate))
        return min(np.ceil(term_months / 12), common.max_term_debts)

    def update(self, year, rate, prices):
        """
        Updates yearly payment and balance.

        Parameters
        ----------
        year : int
            year
        rate : float
            interest rate
        prices : Prices
            instance of the class Prices
        """
        if self.term == 0:
            self.payment, self.balance = 0, 0
        else:
            self.payment = (self.balance*(rate/(1 + rate)) /
                            (1 - (1 / (1 + rate))**self.term))
            new_bal = max(self.balance - self.payment, 0)
            self.balance = new_bal * (1 + rate)
            self.term -= 1

    def reset(self):
        """
        Reset balance and term to initial values.
        """
        self.balance = self.init_balance
        self.term = self.init_term



import pandas as pd
import numpy as np
import requests
from os import path


class table:
    """
    Class computing annuity factors by province, gender, age and birth year.

    Parameters
    ----------
    prov: str
        province; default=qc
    scenario: str
        longevity scenario; default=M (Medium mortality)
    gender: str
        gender; default=males
    web: bool
        True to pull data from internet (option unavailable at the moment), False otherwise; default=False
    """
    def __init__(self, prov='qc', scenario='M', gender='males', web=False):
        self.params = path.join(path.dirname(__file__), 'params')
        quotes = pd.read_excel(self.params+'/quotients.xlsx',
                               sheet_name='DTH_Tx_Scenario_' + scenario,
                               skiprows=3)
        provdict = {'nl': 10,
                    'pe': 11,
                    'ns': 12,
                    'nb': 13,
                    'qc': 24,
                    'on': 35,
                    'mb': 46,
                    'sk': 47,
                    'ab': 48,
                    'bc': 59,
                    'yu': 60,
                    'nt': 61,
                    'nu': 62}
        sexdict = {'males': 1, 'females': 2}
        quotes = quotes[(quotes.CGT == provdict[prov]) &
                        (quotes.Sexe == sexdict[gender])]
        quotes = quotes.drop(columns=[-1, '110+', 'CGT'])
        names = list(quotes.columns.values)
        names[0] = 'year'
        names[1] = 'gender'
        quotes.columns = names
        quotes[['year', 'temp']] = quotes['year'].str.split('-', n=1,
                                                            expand=True)
        # quotes['year'], quotes['temp'] = quotes['year'].str.split('-', 1).str
        quotes = quotes.drop(columns=['temp', 'gender'])
        quotes.set_index('year')
        # transform quotients into rates
        for a in range(0, 110):
            quotes[a] = 2.0*quotes[a] / (2.0-quotes[a])
            quotes[a] = np.where(quotes[a] > 1.0, 1.0, quotes[a])
        quotes = quotes.set_index('year')
        quotes[110] = quotes[109]
        quotes.columns = [str(i) for i in range(0, 111)]
        self.prospect = quotes
        if web is False:
            self.pull_history(prov, gender)
        else:
            self.pull_history_web(prov, gender)
        self.splice()

    # north west territories given saskatchewan retrospective mortality
    # because too short history.
    # newfoundland and labrador given new brunswick for same reason
    def pull_history_web(self, prov='qc', gender='males'):
        """
        Function to get historical mortalilty tables by province and gender.

        Parameters
        ----------
        prov: str
            province; default=qc
        gender: str
            gender; default=males
        """
        maprov = {'qc': 'que',
                  'on': 'ont',
                  'bc': 'bco',
                  'ab': 'alb',
                  'nl': 'nbr',
                  'pe': 'pei',
                  'ns': 'nsc',
                  'nb': 'nbr',
                  'mb': 'man',
                  'sk': 'sas',
                  'nt': 'sas',
                  'yu': 'yuk'}
        file = ('http://www.prdh.umontreal.ca/BDLC/data/' + maprov[prov]
                + '/Mx_1x1.txt')
        r = requests.get(file)
        list_r = r.content.decode().split("\n")
        names = list_r[2].split()
        list_r = list_r[3:]
        r = 0
        for line in list_r:
            list_r[r] = line.split()
            r += 1
        history = pd.DataFrame(data=list_r, index=None, columns=names)
        history.loc[history.Age == '110+', 'Age'] = '110'
        history.dropna(axis=0, inplace=True)
        history['Year'] = history['Year'].astype('int64')
        history['Age'] = history['Age'].astype('int64')
        names = names[2:]

        for n in names:
            history[n] = np.where(history[n] == '.', '1.0', history[n])
            history[n] = history[n].astype('float64')
        if (gender == 'males'):
            history = history[['Year', 'Age', 'Male']]
        else:
            history = history[['Year', 'Age', 'Female']]
        history.columns = ['year', 'age', 'mx']
        history = pd.pivot_table(history, index=['year'], columns=['age'],
                                 values=['mx'])
        history = history['mx']
        history.columns = [str(a) for a in range(0, 111)]
        for a in range(0, 111):
            history[str(a)] = np.where(history[str(a)] > 1.0, 1.0,
                                       history[str(a)])
        self.history = history
        self.history.index = [str(i) for i in range(1921, 2012)]
        self.history.to_excel(self.params + f'/{prov}_{gender}.xlsx')

    def pull_history(self, prov='qc', gender='males'):
        """
        Function to get historical mortalilty tables by province and gender.

        Parameters
        ----------
        prov: str
            province; default=qc
        gender: str
            gender; default=males
        """
        self.history = pd.read_excel(self.params + f'/{prov}_{gender}.xlsx')
        self.history.index = [str(i) for i in range(1921, 2012)]

    def splice(self):
        """
        Function to add survival probabilities for the 2012 birth year and after.
        """
        self.history.loc['2012'] = self.history.loc['2011']
        self.tab = self.history.append(self.prospect, sort=True)

    def compute_annuity_factor(self, byear, agestart, rate):
        """
        Function to compute an annuity factor.

        Parameters
        ----------
        byear : int
            birth year
        agestart : int
            age at which annuity starts
        rate : float
            interest rate

        Returns
        -------
        float
            Annuity factor.
        """
        yrstart = byear + agestart
        prob_surv = 1
        present_value = 1

        for i in range(1, 111 - agestart):
            if yrstart + i < 2063:
                prob_surv *= (1 - self.tab.loc[str(yrstart + i - 1),
                                               str(agestart + i - 1)])
            else:
                prob_surv *= (1 - self.tab.loc['2062', str(agestart + i - 1)])
            present_value += prob_surv / ((1 + rate)**i)
        return present_value



import sys
from CPR import tools
sys.path.insert(1, 'C:/Users/pyann/Dropbox (CEDIA)/srd/Model')
import srd


def compute_after_tax_amount(hh, year, common, prices):
    """
    Compute net nominal return and net amount from withdrawal
    for unregistered assets.

    Parameters
    ----------
    hh: Hhold
        household
    year : int
        year
    common : Common
        instance of the class Common
    prices : Prices
        instance of the class Prices
    """
    nom = tools.create_nom(year, prices)
    hh_srd = file_household_inc_to_tax(hh, year, common, prices)
    hh_srd_0 = file_household(hh, year, common, prices)

    for who, p in enumerate(hh.sp):
        if p.liquidation_to_tax:
            p.liquidation_after_tax = nom(
                hh_srd.sp[who].disp_inc - hh_srd_0.sp[who].disp_inc)
        else:
            p.fin_assets['unreg'].amount_after_tax = nom(
                hh_srd.sp[who].disp_inc - hh_srd_0.sp[who].disp_inc)


def file_household(hh, year, common, prices):
    """
    Files household using SRD.

    Parameters
    ----------
    hh: Hhold
        household
    year : int
        year
    common : Common
        instance of the class Common
    prices : Prices
        instance of the class Prices

    Returns
    -------
    Hhold
        instance of class Hhold after tax
    """
    real = tools.create_real(year, prices)

    hh_srd = []  # persons in household (1 or 2)
    for p in hh.sp:
        p_srd = srd.Person(age=p.age, earn=real(p.earn), rpp=real(p.rpp),
                           cpp=real(p.cpp),
                           net_cap_gains=real(p.net_cap_gains),
                           prev_cap_losses=real(p.prev_cap_losses),
                           div_elig=real(p.div_elig),
                           div_other_can=real(p.div_other_can),
                           othtax=real(p.other_taxable),
                           othntax=real(p.other_non_taxable),
                           inc_rrsp=real(p.inc_rrsp),
                           con_rrsp=real(p.con_rrsp),
                           con_non_rrsp=real(p.con_non_rrsp),
                           asset=real(p.assets))
        hh_srd.append(p_srd)

    if hh.couple:
        hh_srd = srd.Hhold(hh_srd[0], second=hh_srd[1], prov=hh.prov)
    else:
        hh_srd = srd.Hhold(hh_srd[0], prov=hh.prov)
    common.tax.compute(hh_srd)

    return hh_srd


def file_household_inc_to_tax(hh, year, common, prices):
    """
    Files household with taxable income (interests and dividends)
    from unregistered assets.

    Parameters
    ----------
    hh: Hhold
        household
    year : int
        year
    common : Common
        instance of the class Common
    prices : Prices
        instance of the class Prices

    Returns
    -------
    Hhold
        instance of class Hhold after tax
    """
    real = tools.create_real(year, prices)

    hh_srd = []  # persons in household (1 or 2)
    for p in hh.sp:
        inc_div_elig = common.share_div_elig * p.fin_assets['unreg'].inc_div
        inc_int_other_div = (p.fin_assets['unreg'].inc_int
                             + p.fin_assets['unreg'].inc_div - inc_div_elig)

        non_taxable = p.liquidation_non_taxable
        cap_gains = p.liquidation_cap_gains
        cap_losses = common.frac_cap_gains * p.liquidation_cap_losses
        business_exempt = p.liquidation_business_exempt

        p_srd = srd.Person(age=p.age, earn=real(p.earn), rpp=real(p.rpp),
                           cpp=real(p.cpp),
                           net_cap_gains=real(p.net_cap_gains + cap_gains),
                           prev_cap_losses=real(p.prev_cap_losses + cap_losses),
                           cap_gains_exempt=real(business_exempt),
                           div_elig=real(p.div_elig + inc_div_elig),
                           div_other_can=real(p.div_other_can),
                           othtax=real(p.other_taxable + inc_int_other_div),
                           othntax=real(p.other_non_taxable + non_taxable),
                           inc_rrsp=real(p.inc_rrsp),
                           con_rrsp=real(p.con_rrsp),
                           con_non_rrsp=real(p.con_non_rrsp),
                           asset=real(p.assets))
        hh_srd.append(p_srd)

    if hh.couple:
        hh_srd = srd.Hhold(hh_srd[0], second=hh_srd[1], prov=hh.prov)
    else:
        hh_srd = srd.Hhold(hh_srd[0], prov=hh.prov)
    common.tax.compute(hh_srd)
    return hh_srd

def get_gis_oas_allowances(hh, hh_tax, year, prices):
    """
    Computes nominal GIS, OAS and allowance benefits

    Parameters
    ----------
    hh: Hhold
        household
    hh_tax : Hhold
        household
    year : int
        year
    prices : Prices
        instance of the class Prices
    """
    nom_2018 = tools.create_nom(year, prices)

    for who, p in enumerate(hh.sp):
        p.inc_gis = nom_2018(hh_tax.sp[who].inc_gis)
        p.inc_oas = nom_2018(hh_tax.sp[who].inc_oas)
        p.allow_surv = nom_2018(hh_tax.sp[who].allow_surv)
        p.allow_couple = nom_2018(hh_tax.sp[who].allow_couple)



import os
import pickle
import pandas as pd
import numpy as np

import srpp

from CPR import life
from CPR import tools

import sys
sys.path.insert(1, 'C:/Users/pyann/Dropbox (CEDIA)/srd/Model')
import srd

module_dir = os.path.dirname(os.path.dirname(__file__))
path_params = '/CPR/data/params/'
path_factors = '/CPR/data/precomputed/'


class CommonParameters:
    """
    This class sets and contains the parameters common to all households.
    
    Parameters
    ----------
    nsim: int
        number of simulations
    non_stochastic: bool
        True if non stochastic simulation, False otherwise
    extra_params: dict
        dictionary of extra parameters
    """
    def __init__(self, nsim, non_stochastic, extra_params):

        self.nsim = nsim
        self.non_stochastic = non_stochastic
        for file in ['common_params.csv', 'user_options.csv']:
            tools.add_params_as_attr(self, module_dir + path_params + file)
        tools.change_params(self, extra_params)

        self.tax = srd.tax(year=self.base_year)

        self.rules_cpp = srpp.rules()
        self.rules_qpp = srpp.rules(qpp=True)

        for name in ['rrsp', 'tfsa']:
            self.set_limits(name)

        self.d_ympe = self.prepare_ympe()
        self.d_perc_cpp = self.prepare_cpp()
        self.d_perc_rrif = tools.get_params(
            module_dir + path_params + 'rrif_rates.csv', numerical_key=True)

    def set_limits(self, name):
        """
        Set contributions limits for RRSP and TFSA.

        Parameters
        ----------
        name: str
            RRSP or TFSA
        """
        d = {}
        for year in range(self.base_year, self.base_year + self.future_years):
            try:
                d[year] = getattr(self, f'{name}_limit_{year}')
            except:
                d[year] = round(d[year-1]
                                * (1+getattr(self, f'gr_{name}_limit')))
        setattr(self, f'd_{name}_limit', d)

    def prepare_ympe(self):
        """"
        Pre-reform ympe used to adjust DB benefits for CPP
        """
        d_ympe = tools.get_params(module_dir + path_params + 'ympe.csv',
                                  numerical_key=True)
        for year in range(max(d_ympe.keys()) + 1,
                          self.base_year + self.future_years):
            d_ympe[year] = round(d_ympe[year-1] * (1 + self.gr_ympe))
        return d_ympe

    def prepare_cpp(self):
        """
        Set percentages for cpp/qpp benefits.
        """
        d_perc_cpp = {year: getattr(self, f'perc_cpp_{year}')
                      for year in range(2018, 2023)}
        d_perc_cpp.update(
            {year: self.perc_cpp_2023
             for year in range(2023, self.base_year + self.future_years)})
        return d_perc_cpp


class Prices:
    """
    This class computes the times series for asset returns,
    interest rates on debt, deterministic wage profiles,
    housing price growth rate and price/rent ratio.
    
    Parameters
    ----------
    common: Common
        instance of the class Common
    extra_params: dict
        dictionary of extra parameters
    """
    def __init__(self, common, extra_params):
        tools.add_params_as_attr(self, module_dir + path_params + 'prices.csv')
        tools.change_params(self, extra_params)
        np.random.seed(self.seed)

        for asset in ['bills', 'bonds', 'equity', 'business']:
            ret = self.simulate_ret(asset, common)
            setattr(self, f'ret_{asset}', ret)

        self.ret_housing, self.price_rent_ratio = self.simulate_housing(common)

        self.d_infl_factors = self.prepare_inflation_factors(common)
        self.d_interest_debt = self.simulate_interest_debt()
        self.d_diff_log_wages = self.attach_diff_log_wages()

        if common.recompute_factors:
            self.d_factors = self.initialize_factors()
        else:
            with open(module_dir + path_factors + 'd_factors.pickle', 'rb') as file:
                self.d_factors = pickle.load(file)

    def simulate_ret(self, asset, common):
        """
        Simulate N series of length T nominal returns distributed lognormally
        with autocorrelation rho.

        Parameters
        ----------
        asset: str
            type of asset
        common: Common
            instance of the class Common

        Returns
        -------
        numpy.array:
            Array of nominal returns
        """
        r = np.empty((common.future_years, common.nsim))
        r[0, :] = getattr(self, f'ret_{asset}_2018')
        mu = getattr(self, f'mu_{asset}')
        rho = getattr(self, f'rho_{asset}')
        sigma = getattr(self, f'sig_{asset}')

        alpha, sig_eps = self.compute_params_process(mu, rho, sigma)

        if common.non_stochastic:
            for t in range(1, common.future_years):
                r[t, :] = np.exp(alpha) * (1+r[t-1, :])**rho \
                    * np.exp(sig_eps**2/2) - 1
        else:
            eps = np.random.normal(loc=0, scale=sig_eps,
                                   size=(common.future_years, common.nsim))
            for t in range(1, common.future_years):
                r[t, :] = np.exp(alpha) * (1+r[t-1, :])**rho \
                    * np.exp(eps[t, :]) - 1
        return (1+r) * (1 + self.inflation_rate) - 1

    def compute_params_process(self, mu, rho, sigma):
        """        
        Convert arithmetic mean mu and volatility sigma of the returns
        and autocorrelation rho of the log returns into
        :math:`\\alpha`, :math:`\\rho` and :math:`\\sigma_{\\epsilon}` of the process:
        
        :math:`\\ln(1+r_t) = \\alpha + \\rho * \\ln(1+r_{t-1}) + \\epsilon`,
        where :math:`\\epsilon \\sim N(0, \\sigma_{\\epsilon})`.

        Parameters
        ----------
        mu: float
            arithmetic mean
        rho: float
            autocorrelation :math:
        sigma: float
            standard deviation

        Returns
        -------
        float:
            AR(1) coefficient (:math:`\\alpha`)
        float:
            Standard deviation of error term (:math:`\\sigma_{\\epsilon}`)
        """
        m, v = (1+mu), sigma**2
        sig_lognorm = np.sqrt(np.log(1 + v / m**2))
        mu_lognorm = np.log(m / np.sqrt(1 + v / m**2))
        alpha = (1-rho) * mu_lognorm
        sig_eps = np.sqrt(1 - rho**2) * sig_lognorm
        return alpha, sig_eps

    def simulate_housing(self, common):
        """
        Simulate series of nominal housing price growth (in :math:`\\ln(1+r)` form)
        and price-rent ratio.

        Parameters
        ----------
        common: Common
            instance of the class Common

        Returns
        -------
        numpy.array:
            Array of nominal housing price growth
        numpy.array:
            Array of price-rent ratios
        """
        r = np.empty((common.future_years, common.nsim))
        r[0, :] = self.ret_housing_2018
        rho_r = self.rho_housing
        alpha_r, sig_eps_r = self.compute_params_process(
            self.mu_housing, rho_r, self.sig_housing)

        ratio = np.empty_like(r)
        ratio[0, :] = self.price_rent_2018
        rho_ratio = self.rho_price_rent
        alpha_ratio = self.mu_price_rent * (1 - rho_ratio)
        sig_eps_ratio = np.sqrt(1 - rho_ratio**2) * self.sig_price_rent

        if common.non_stochastic:
            for t in range(1, common.future_years):
                r[t, :] = np.exp(alpha_r) * (1 + r[t-1, :])**rho_r \
                    * np.exp(sig_eps_r**2/2) - 1
                ratio[t, :] = alpha_ratio + rho_ratio * ratio[t-1, :]
        else:
            cov = sig_eps_r * sig_eps_ratio * self.corr_housing_price_rent
            m_cov = np.array([[sig_eps_r**2, cov],
                              [cov, sig_eps_ratio**2]])
            eps = np.random.multivariate_normal(
                [0, 0], m_cov, size=(common.future_years, common.nsim))
            eps_r, eps_ratio = eps[:, :, 0], eps[:, :, 1]

            for t in range(1, common.future_years):
                r[t, :] = np.exp(alpha_r) * (1+r[t-1, :])**rho_r \
                    * np.exp(eps_r[t, :]) - 1
                ratio[t, :] = alpha_ratio + rho_ratio * ratio[t-1, :] \
                    + eps_ratio[t, :]
        return (1+r) * (1 + self.inflation_rate) - 1, ratio

    def prepare_inflation_factors(self, common):
        """
        Compute inflation factors with base year 2018.

        Parameters
        ----------
        common: Common
            instance of the class Common

        Returns
        -------
        dict:
            Dictionary of inflation factors for each year
        """
        start_year = common.base_year - common.past_years
        end_year = common.base_year + common.future_years
        # future inflation
        d_infl_factors = {
            year: (1 + self.inflation_rate)**(year-common.base_year)
            for year in range(common.base_year, end_year)}
        # past inflation
        d_inflation = tools.get_params(
            module_dir + path_params + 'inflation.csv', numerical_key=True)
        for year in reversed(range(start_year, common.base_year)):
            d_infl_factors[year] = (d_infl_factors[year + 1]
                                    / (1 + d_inflation[year]))
        return d_infl_factors

    def simulate_interest_debt(self):
        """
        Creates N series of yearly nominal interest rate of length T
        for each type of debt

        Returns
        -------
        dict:
            Dictionary of interest rates by type of debt and year
        """
        mix_fee_debts = pd.read_csv(
            module_dir + path_params + "mix_fee_debt.csv",
            usecols=list(range(5)), index_col=0).to_dict('index')
        # monthly rates:
        d_interest = {}
        for debt in mix_fee_debts:
            d_interest[debt] = (mix_fee_debts[debt]['bills']*self.ret_bills
                                + mix_fee_debts[debt]['bonds']*self.ret_bonds
                                + mix_fee_debts[debt]['fee'])
            d_interest[debt][0] = mix_fee_debts[debt]['value_2018']

        return d_interest

    def attach_diff_log_wages(self):
        """
        Creates a dictionary of differences in log wages by education and age.

        Returns
        -------
        dict:
            Dictionary of difference in log wages by education and age
        """
        diff_log_wages = pd.read_csv(
            module_dir + path_params + 'diff_log_wage.csv', index_col=0)
        d_diff_log_wages = {}
        for degree in diff_log_wages.columns:
            d_diff_log_wages[degree] = np.cumsum(
                diff_log_wages[degree].values + np.log(1 + self.gr_rate_wage))
        return d_diff_log_wages

    def initialize_factors(self):
        """
        This function creates an instance of life.table
        by gender and province.

        Returns
        -------
        dict:
            dictionary of annuity factors by gender and provinces
        """
        l_sex = ['male', 'female']
        l_prov = ['qc', 'on', 'ab', 'bc', 'sk', 'ns', 'nb', 'mb', 'pe', 'nl']
        d_factors = {}
        for s in l_sex:
            d_factors[s] = {}
            for p in l_prov:
                d_factors[s][p] = life.table(prov=p, scenario='M',
                                             gender=s+'s')
        with open(module_dir + path_factors + 'd_factors.pickle', 'wb') as file:
            pickle.dump(d_factors, file)
        return d_factors



import os
import pandas as pd
import numpy as np
from CPR import macro

module_dir = os.path.dirname(os.path.dirname(__file__))


def get_dataset():
    """
    Function that returns a dataframe of synthetic data.

    Returns
    -------
    pandas.core.frame.DataFrame
        Dataframe of synthetic data.
    """
    return pd.read_csv(module_dir + '/CPR/data/inputs/synth_inputs.csv',
                       index_col=0)


class Results:
    """
    This class prepares the results.

    Parameters
    ----------
    input_data: pandas.core.frame.DataFrame
        dataframe of inputs
    output: pandas.core.frame.DataFrame
        dataframe of outputs
    common: Common
        instance of the class Common
    prices: Prices
        instance of the class Prices
    extra_params: dict
        dictionary of extra parameters
    """
    def __init__(self, input_data, output, common, prices, extra_params):
        self.input = input_data
        self.output = output
        self.common = common
        self.prices = prices
        self.extra_params = extra_params

    def summarize(self):
        """
        Function to summarize the simulation.
        """
        if self.common.non_stochastic is True:
            print('\nDeterministic Model')
        else:
            print(f'\nStochastic Model, {self.common.nsim} simulations')

        if self.extra_params != {}:
            print('\nExtra parameters:')
            for k, v in self.extra_params.items():
                print(f'{k}: {v}')


    def merge(self, add_index=True):
        """
        Function to merge input and output variables to create a database.

        Parameters
        ----------
        add_index : bool, optional
            add index to database, True by default
        """
        self.df_merged = self.input.merge(self.output, how='left',
                                          left_index=True, right_on='hh_index')
        if add_index:
            if self.common.nsim == 1:
                self.df_merged.set_index('hh_index', inplace=True)
            else:
                self.df_merged.set_index(['hh_index', 'sim'], inplace=True)


    def check_preparedness(self, factor_couple=2, cons_floor=100,
                           d_cutoffs={20: 80, 100: 65}):
        """
        Function that introduces a consumption floor, computes RRI (NaN if consumption before and after retirement below cons_floor), and checks preparedness for retirement.

        Parameters
        ----------
        factor_couple : int, optional
            factor to normalize income for couple, by default 2
        cons_floor : int, optional
            consumption floor, by default 100
        d_cutoffs : dict, optional
            cutoffs, by default {20: 80, 100: 65}
        """
        if not hasattr(self, 'df_merged'):
            self.merge()

        # adjusts consumption
        self.df_merged[['cons_bef', 'cons_after']] = \
            self.df_merged[['cons_bef', 'cons_after']].clip(lower=cons_floor)
        # compute RRI
        self.df_merged['rri'] = \
            self.df_merged.cons_after / self.df_merged.cons_bef * 100
        mask = ((self.df_merged.cons_bef == cons_floor) &
                (self.df_merged.cons_after == cons_floor))
        self.df_merged.loc[mask, 'rri'] = np.nan

        # compute cutoffs
        def compute_hh_wage(row):
            """
            Function to compute a normalized wage for the household.

            Parameters
            ----------
            row : pandas.core.series.Series
                row of pandas.core.frame.DataFrame

            Returns
            -------
            float
                Normalized wage.
            """
            if row['couple']:
                return (row['init_wage'] + row['s_init_wage']) / factor_couple
            return row['init_wage']

        self.df_merged['hh_init_adj_wage'] = \
            self.df_merged.apply(compute_hh_wage, axis=1)

        l_quantiles = [k/100 for k in d_cutoffs.keys()][:-1]
        l_cutoffs = self.df_merged.hh_init_adj_wage.quantile(l_quantiles)
        d_cutoffs_by_cat = {i: v for i, (k, v) in enumerate(d_cutoffs.items())}

        # apply cutoffs to households inital wages
        self.df_merged['category'] = \
            np.searchsorted(l_cutoffs, self.df_merged.hh_init_adj_wage)
        self.df_merged['rri_cutoff'] = \
            self.df_merged.category.replace(d_cutoffs_by_cat)
        self.df_merged.drop(columns='category', inplace=True)
        self.df_merged['prepared'] = \
            self.df_merged.rri >= self.df_merged.rri_cutoff



import heapq
import numpy as np
from CPR import tools


class ContributionRoom:
    """
    This class manages contribution room for TFSAs and RRSPs.

    All amounts are nominal.

    Parameters
    ----------
    init_room_rrsp: float
        initial RRSP contribution room available
    init_room_tfsa: float
        initial TFSA contribution room available
    """
    def __init__(self, init_room_rrsp, init_room_tfsa):
        self.init_room_rrsp = init_room_rrsp
        self.init_room_tfsa = init_room_tfsa
        self.reset()

    def compute_contributions(self, p, year, common, prices):
        """
        Function to update contribution room for RRSPs and TFSAs, using the 2 other functions below (which themselves call the other functions of the class).

        Parameters
        ----------
        p: Person
            instance of the class Person
        year : int
            year
        common : Common
            instance of the class Common
        prices : Prices
            instance of the class Prices
        """
        self.update_rrsp_room(p, year, common)
        self.update_tfsa_room(p, year, common)
        self.extra_contrib_rrsp = 0
        if (p.replacement_rate_db > 0) & (year < common.base_year
                                          + common.max_years_db):
            self.adjust_db_contributions(p, year, common)
        if hasattr(p, 'rpp_dc'):
            self.adjust_dc_contributions(p, year)
        for acc in set(p.fin_assets) & set(('rrsp', 'other_reg')):
            self.adjust_rrsp_contributions(acc, p, year)

        self.adjust_tfsa_contributions(p, year, common, prices)

        self.adjust_unreg_contributions(p, year)

    def update_rrsp_room(self, p, year, common):
        """
        Function to update RRSP contribution room.

        Parameters
        ----------
        p: Person
            instance of the class Person
        year : int
            year
        common : Common
            instance of the class Common
        """
        if p.age > common.max_age_no_rrif:
            self.room_rrsp = 0
        else:
            self.room_rrsp += min(common.perc_rrsp * p.d_wages[year-1],
                                  common.d_rrsp_limit[year])

    def update_tfsa_room(self, p, year, common):
        """
        Function to update TFSA contribution room.

        Parameters
        ----------
        p: Person
            instance of the class Person
        year : int
            year
        common : Common
            instance of the class Common
        """
        self.room_tfsa += common.d_tfsa_limit[year]
        self.room_tfsa += p.fin_assets['tfsa'].withdrawal

    def adjust_db_contributions(self, p, year, common):
        """
        Function to adjust RRSP contribution room to DB RPP contributions.

        Parameters
        ----------
        p: Person
            instance of the class Person
        year : int
            year
        common : Common
            instance of the class Common
        """
        cpp_offset = (common.perc_cpp_2018
                      * min(p.d_wages[year-1], common.d_ympe[year-1])
                      / common.max_years_db)
        benefits_earned = (p.replacement_rate_db * p.d_wages[year-1]
                           / common.max_years_db - cpp_offset)
        pension_credit = max(0, common.db_benefit_multiplier * benefits_earned
                             - common.db_offset)
        self.room_rrsp -= min(pension_credit, self.room_rrsp)

    def adjust_dc_contributions(self, p, year):
        """
        Function to adjust RRSP contribution room to DC RPP contributions.

        If contribution room is insufficent for the intended/planned RRSP contributions, the "excess" contributions are channeled to TFSA.

        Parameters
        ----------
        p: Person
            instance of the class Person
        year : int
            year
        """
        contrib = p.rpp_dc.contrib_rate * p.d_wages[year]
        if contrib <= self.room_rrsp:
            p.rpp_dc.contribution = contrib
            self.room_rrsp -= contrib
        else:
            p.rpp_dc.contribution = self.room_rrsp
            self.extra_contrib_rrsp += contrib - self.room_rrsp
            self.room_rrsp = 0
        self.adjust_employees_contributions(p)

    def adjust_employees_contributions(self, p):
        """
        Function to compute employee contributions to DC RPPs (later used to caculate taxes).

        Parameters
        ----------
        p: Person
            instance of the class Person
        """
        if p.rpp_dc.contrib_rate == 0:
            p.contrib_employee_dc = 0
        else:
            p.contrib_employee_dc = (p.rate_employee_dc /
                                     p.rpp_dc.contrib_rate
                                     * p.rpp_dc.contribution)

    def adjust_rrsp_contributions(self, acc, p, year):
        """
        Function to adjust RRSP contribution for contributions other than to RPPs.

        If contribution room is insufficent for the intended/planned RRSP contributions, the "excess" contributions are channeled to TFSA.

        Parameters
        ----------
        acc : str
            type of account
        p: Person
            instance of the class Person
        year : int
            year
        """
        contrib = p.fin_assets[acc].contrib_rate * p.d_wages[year]
        if contrib <= self.room_rrsp:
            p.fin_assets[acc].contribution = contrib
            self.room_rrsp -= contrib
        else:
            p.fin_assets[acc].contribution = self.room_rrsp
            self.extra_contrib_rrsp += contrib - self.room_rrsp
            self.room_rrsp = 0

    def adjust_tfsa_contributions(self, p, year, common, prices):
        """
        Function to adjust TFSA contribution room to TFSA contributions (including "excess" DC RPP and RRSP contributions).

        If contribution room is insufficent for the intended/planned TFSA contributions, the "excess" contributions are channeled to unregistered accounts.

        Parameters
        ----------
        p: Person
            instance of the class Person
        year : int
            year
        common : Common
            instance of the class Common
        prices : Prices
            instance of the class Prices
        """
        self.extra_contrib_tfsa = 0
        contrib = p.fin_assets['tfsa'].contrib_rate * p.d_wages[year]
        contrib += self.extra_contrib_rrsp
        if p.age > common.max_age_no_rrif:
            contrib += self.adjust_rrif(p, year, common, prices)

        if contrib <= self.room_tfsa:
            p.fin_assets['tfsa'].contribution = contrib
            self.room_tfsa -= contrib
        else:
            p.fin_assets['tfsa'].contribution = self.room_tfsa
            self.extra_contrib_tfsa += contrib - self.room_tfsa
            self.room_tfsa = 0

    def adjust_rrif(self, p, year, common, prices):
        """
        Function to adjust DC RPP and RRSP accounts to mandatory RRIF withdrawals (RRIFs are not separately modelled in this version).

        Parameters
        ----------
        p: Person
            instance of the class Person
        year : int
            year
        common : Common
            instance of the class Common
        prices : Prices
            instance of the class Prices

        Returns
        -------
        float
            Nominal amount that needs to be withdrawn.
        """
        rrif_transfer_real = 0
        for acc in set(p.fin_assets) & set(('rpp_dc', 'rrsp', 'other_reg')):
            rrif_transfer_real += p.fin_assets[acc].rrif_withdrawal(p, common)
        return rrif_transfer_real * prices.d_infl_factors[year]

    def adjust_unreg_contributions(self, p, year):
        """
        Function to adjust contributions to unregistered accounts for "excess" TFSA contributions channeled to them.

        Parameters
        ----------
        p: Person
            instance of the class Person
        year : int
            year
        """
        contrib = p.fin_assets['unreg'].contrib_rate * p.d_wages[year]
        contrib += self.extra_contrib_tfsa
        p.fin_assets['unreg'].contribution = contrib

    def reset(self):
        """
        Reset RRSP and TFSA contribution rooms to their inital values.
        """
        self.room_rrsp = self.init_room_rrsp
        self.room_tfsa = self.init_room_tfsa


class FinAsset:
    """
    This class manages registered accounts. All amounts are nominal.

    Parameters
    ----------
    p: Person
        instance of the class Person
    hh: Hhold
        household
    acc: str
        type of account
    """
    def __init__(self, p, hh, acc):
        self.init_balance = getattr(p, f'bal_{acc}')
        self.contrib_rate = getattr(p, f'cont_rate_{acc}')
        self.init_desired_withdrawal_real = getattr(p, f'withdrawal_{acc}')
        self.mix_bills = hh.mix_bills
        self.mix_bonds = hh.mix_bonds
        self.mix_equity = hh.mix_equity
        self.fee = hh.fee
        self.reset()

    def update(self, d_returns, year, common, prices):
        """
        Function to update the balance to account for contributions, withdrawals, and returns.

        Parameters
        ----------
        d_returns : dict
            dictionary of returns
        year : int
            year
        common : Common
            instance of the class Common
        prices : Prices
            instance of the class Prices
        """
        self.nom, self.real = tools.create_nom_real(year, prices)
        self.balance *= 1 + self.rate(d_returns, year)
        self.balance += self.contribution
        self.desired_withdrawal = self.nom(self.desired_withdrawal_real)
        self.withdrawal = min(self.balance, self.desired_withdrawal)
        self.balance -= self.withdrawal

    def rate(self, d_returns, year):
        """
        Function to compute the rate of return given the mix of assets in the account.

        Parameters
        ----------
        d_returns : dict
            dictionary of returns
        year : int
            year

        Returns
        -------
        float
            Rate of return (net of fees).
        """
        return (self.mix_bills*d_returns['bills'][year]
                + self.mix_bonds*d_returns['bonds'][year]
                + self.mix_equity*d_returns['equity'][year] - self.fee)

    def rrif_withdrawal(self, p, common):
        """
        Function to manage mandatory RRIF withdrawals.

        Parameters
        ----------
        p: Person
            instance of the class Person
        common : Common
            instance of the class Common

        Returns
        -------
        float
            Amount of mandatory withdrawal.
        """
        if self.balance > 0:
            extra_withdrawal_real = max(
                0, common.d_perc_rrif[p.age] * self.real(self.balance)
                - self.desired_withdrawal_real)
            self.desired_withdrawal_real += extra_withdrawal_real
        else:
            extra_withdrawal_real = 0

        return extra_withdrawal_real

    def liquidate(self):
        """
        Function to liquidate an account, setting balance, contributions and withdrawals to zero.
        
        Returns
        -------
        float
            Amount from liquidation (before taxes).
        """
        value = self.balance
        self.balance, self.contribution = 0, 0
        return value

    def reset(self):
        """
        Reset the balance and withdrawal to its initial balance.
        """
        self.balance = self.init_balance
        self.desired_withdrawal_real = self.init_desired_withdrawal_real
        self.withdrawal = 0  # to adjust contribution space tfsa first period


class UnregAsset:
    """
    This class manages unregistered accounts.

    All amounts are nominal.

    Parameters
    ----------
    p: Person
        instance of the class Person
    hh: Hhold
        household
    prices: Prices
        instance of the class Prices
    """
    def __init__(self, p, hh, prices):
        self.init_balance = p.bal_unreg
        self.contrib_rate = p.cont_rate_unreg
        self.desired_withdrawal = p.withdrawal_unreg
        self.init_cap_gains = p.cap_gains_unreg
        self.init_realized_losses = p.realized_losses_unreg
        self.mix_bills = hh.mix_bills
        self.mix_bonds = hh.mix_bonds
        self.mix_equity = hh.mix_equity
        self.fee = hh.fee
        self.fee_equity = hh.fee_equity
        self.ret_dividends = prices.ret_dividends
        self.reset()

    def update(self, d_returns, year, common, prices):
        """
        Function to update the balance for contributions, withdrawals and returns.

        Parameters
        ----------
        d_returns : dict of float
            dictionary of returns
        year : int
            year
        common : Common
            instance of the class Common
        prices : Prices
            instance of the class Prices
        """
        self.nom = tools.create_nom(year, prices)
        self.compute_income(d_returns, year)
        self.update_balance()
        self.prepare_withdrawal()
        self.adjust_income()
        self.withdrawal_cap_losses = self.adjust_cap_losses(
            self.withdrawal_cap_gains)
        self.adjust_final_balance()
        self.amount_after_tax = 0
        self.income_to_tax = bool(self.inc_div + self.inc_int)

    def compute_income(self, d_returns, year):
        """
        Function to compute new capital gains and taxable income (dividends and interests).

        Parameters
        ----------
        d_returns : dict of float
            dictionary of returns
        year : int
            year
        """
        # to simplify, we assume constant dividend rate and fees paid out
        # of capital gains.
        self.inc_div = self.balance * self.mix_equity * self.ret_dividends
        self.inc_cap_gains = (
            self.balance * self.mix_equity
            * (d_returns['equity'][year] - self.fee_equity)
            - self.inc_div)
        self.inc_int = (self.balance * self.rate(d_returns, year)
                        - self.inc_div - self.inc_cap_gains)

    def rate(self, d_returns, year):
        """
        Function that computes the rate of return given the mix of assets in account.

        Parameters
        ----------
        d_returns : dict of float
            dictionary of returns
        year : int
            year

        Returns
        -------
        float
            Rate of return (net of fees).
        """
        return (self.mix_bills*d_returns['bills'][year] +
                self.mix_bonds*d_returns['bonds'][year] +
                self.mix_equity*d_returns['equity'][year] - self.fee)

    def update_balance(self):
        """
        Function to update to balance and separate between non-taxable funds (i.e. previous post-tax balance), capital gains, dividends, and interests.
        """
        self.non_taxable += self.amount_after_tax
        self.non_taxable += self.contribution
        self.cap_gains += self.inc_cap_gains
        self.balance = (self.non_taxable + self.cap_gains + self.inc_int
                        + self.inc_div)

    def prepare_withdrawal(self):
        """
        Function to separate a withdrawal from the balance at the end of the period, and separately identify non-taxable funds, capital gains, dividends, and interests.
        """
        self.withdrawal = min(self.nom(self.desired_withdrawal), self.balance)
        self.share_withdrawal = self.withdrawal / (self.balance + 1e-12)
        self.withdrawal_non_tax = self.share_withdrawal * self.non_taxable
        self.withdrawal_cap_gains = self.share_withdrawal * self.cap_gains
        self.withdrawal_div = self.share_withdrawal * self.inc_div
        self.withdrawal_int = self.share_withdrawal * self.inc_int

    def adjust_income(self):
        """
        Function to adjust investment income (dividends and interests) for withdrawals.
        """
        self.inc_div *= 1 - self.share_withdrawal
        self.inc_int *= 1 - self.share_withdrawal

    def adjust_cap_losses(self, realized_cap_gains):
        """
        Function to compute capital losses from previous years used for deduction 
        against capital gains, and adjust realized capital losses accordingly.

        Parameters
        ----------
        realized_cap_gains : float
            capital gains

        Returns
        -------
        float
            Capital losses (to be deducted).
        """
        if realized_cap_gains > 0:
            used_cap_losses = min(realized_cap_gains, self.realized_losses)
            self.realized_losses -= used_cap_losses
        else:
            self.realized_losses += -realized_cap_gains
            used_cap_losses = 0

        return used_cap_losses

    def adjust_final_balance(self):
        """
        Function that adjusts the final balance for withdrawals, dividends and interests.
        """
        self.non_taxable *= (1 - self.share_withdrawal)
        self.cap_gains *= (1 - self.share_withdrawal)
        self.balance = int(self.non_taxable + self.cap_gains)
        # int to avoid approx errors

    def liquidate(self):
        """
        Function to liquidate the account and adjust capital losses.
        """
        self.liquidation_non_taxable = self.non_taxable
        self.liquidation_cap_gains = self.cap_gains
        self.balance, self.non_taxable, self.cap_gains = 0, 0, 0
        self.inc_int, self.inc_div = 0, 0

        self.liquidation_cap_losses = self.adjust_cap_losses(
            self.liquidation_cap_gains)

    def reset(self):
        """
        Function to reset the balance, capital gains, and withdrawals to their initial values.
        """
        self.balance = self.init_balance
        self.cap_gains = self.init_cap_gains
        self.non_taxable = self.balance - self.cap_gains
        self.realized_losses = self.init_realized_losses
        self.amount_after_tax = 0
        self.liquidation_non_taxable = 0
        self.liquidation_cap_gains = 0
        self.liquidation_cap_losses = 0


class RppDC(FinAsset):
    """
    This class manages DC RPPs.

    All amounts are nominal.

    Parameters
    ----------
    p: Person
        instance of the class Person
    common: Common
        instance of the class Common
    """
    def __init__(self, p, common):
        self.init_balance = p.init_dc
        self.contrib_rate = p.rate_employee_dc + p.rate_employer_dc
        self.init_desired_withdrawal_real = 0
        self.mix_bills = common.mix_bills_rpp
        self.mix_bonds = common.mix_bonds_rpp
        self.mix_equity = common.mix_equity_rpp
        self.fee = common.fee_rpp
        self.reset()


class RppDB:
    """
    This class manages DB RPPs.

    All amounts are nominal.

    Parameters
    ----------
    p: Person
        instance of the class Person
    """
    def __init__(self, p):
        self.init_rate_employee_db = p.rate_employee_db
        self.replacement_rate_db = p.replacement_rate_db
        self.income_previous_db = p.income_previous_db
        self.reset()

    def compute_benefits(self, p, common):
        """
        Function that computes RPP DB benefits and adjusts them for CPP/QPP integration.

        If RPP benefits are smaller than CPP/QPP benefits, RPP benefits are set to zero once the receipt of CPP/QPP retirement benefits begins.

        Parameters
        ----------
        p: Person
            instance of the class Person
        common : Common
            instance of the class Common
        """
        if self.replacement_rate_db > 0:
            n = common.n_best_wages_db
            l_wages = [wage for wage in p.d_wages.values()]
            n_highest_wages = sorted(l_wages, reverse=True)[:n]
            self.mean_best_wage = sum(n_highest_wages) / n

            self.adjust_for_penalty(p, common)

            if p.age >= common.official_ret_age:
                self.benefits -= min(self.compute_cpp_adjustment(p, common),
                                     self.benefits_current_empl)

        if self.income_previous_db > 0:
            self.benefits += self.income_previous_db

    def adjust_for_penalty(self, p, common):
        """
        Function to compute a penalty for individuals who begin to receive DB RPP benefits "early", i.e. before they accumulate the maximum number of years of service, if they are younger than the age at which benefits can start without penalty (the "early retirement age").

        By default, the penalty applies to those who begin RPP benefits receipt before reaching 35 years of service and before age 62. These values can be modified.

        Parameters
        ----------
        p: Person
            instance of the class Person
        common : Common
            instance of the class Common
        """
        years_service = p.replacement_rate_db / common.perc_year_db
        cond0 = years_service < common.max_years_db
        cond1 = p.ret_age < common.db_ret_age_no_penalty
        if cond0 and cond1:
            years_early_db = common.official_ret_age - p.ret_age
        else:
            years_early_db = 0

        self.benefits_current_empl = (
            self.replacement_rate_db * self.mean_best_wage
            * (1 - years_early_db * common.db_penalty_early_ret))
        self.benefits += self.benefits_current_empl

    def compute_cpp_adjustment(self, p, common):
        """
        Function to compute an adjustment (reduction) to DB RPP benefits to account for CPP/QPP integration.

        Parameters
        ----------
        p: Person
            instance of the class Person
        common : Common
            instance of the class Common

        Returns
        -------
        float
            Amount of benefit adjustment for CPP/QPP integration.
        """
        years_db = int(self.replacement_rate_db / common.perc_year_db)

        if years_db <= p.ret_year - common.base_year:
            mean_wage = np.mean([min(p.d_wages[common.base_year + t],
                                     common.d_ympe[common.base_year + t])
                                 for t in range(years_db)])
        else:
            mean_wage = np.mean([min(p.d_wages[p.ret_year - t],
                                     common.d_ympe[p.ret_year - t])
                                 for t in range(1, years_db + 1)])

        return (common.perc_cpp_2018 * years_db / common.max_years_db
                * mean_wage)

    def reset(self):
        """
        Function that resets the benefits and contribution rates to their initial values.
        """
        self.benefits = 0
        self.rate_employee_db = self.init_rate_employee_db


class RealAsset:
    """
    This class manages housing and residences.

    All amounts are nominal.

    Parameters
    ----------
    d_hh: dict of float
        dictionary containing values of residences
    resid: str
        primary or secondary residence
    """
    def __init__(self, d_hh, resid):
        self.init_balance = d_hh[resid]
        self.price = d_hh[f'price_{resid}']
        self.reset()

    def update(self, growth_rates, year, prices):
        """
        Function to update the balance (residence values) for growth in price.

        Parameters
        ----------
        growth_rates : dict
            nominal return to housing (capital gains)
        year : int
            year
        prices : Prices
            instance of the class Prices
        """
        self.balance *= 1 + growth_rates[year]

    def liquidate(self):
        """
        Function to liquidate the asset, compute capital gains if they apply, and set balance to zero.

        Returns
        -------
        float
            Amount from asset liquidation (before taxes).
        """
        self.liquidation_non_taxable = self.price
        self.liquidation_cap_gains = self.balance - self.price
        self.balance = 0

    def reset(self):
        """
        Function to reset residence value to its initial value and set capital gains to zero.
        """
        self.balance = self.init_balance
        self.cap_gains = 0

    def impute_rent(self, hh, year, common, prices):
        """
        Function to compute imputed rent.

        Parameters
        ----------
        hh: Hhold
            household
        year : int
            year
        prices : Prices
            instance of the class Prices
        """
        hh.imputed_rent = ((1 - common.downsize) 
                           * hh.residences['first_residence'].balance
                           / prices.d_price_rent_ratio[year])


class Business:
    """
    This class manages a business (as an asset owned by the houshehold).

    All amounts are nominal.

    Parameters
    ----------
    d_hh: dict of float
        dictionary containing value of business
    """
    def __init__(self, d_hh):
        self.init_balance = d_hh['business']
        self.price = d_hh['price_business']
        self.reset()

    def update(self, d_business_returns, year, prices):
        """
        Function to update the balance and dividends.

        Parameters
        ----------
        d_business_returns : dict
            returns on business assets
        year : int
            year
        prices : Prices
            instance of the class Prices
        """
        self.dividends_business = self.balance * prices.ret_dividends
        self.balance *= 1 + d_business_returns[year] - prices.ret_dividends

    def liquidate(self, common):
        """
        Function to liquidate account (business assets), compute capital gains, and set balance to zero.

        Parameters
        ----------
        common : Common
            instance of the class Common

        Returns
        -------
        float
            Selling price of the business assets.
        """
        self.liquidation_non_taxable = self.price
        self.liquidation_cap_gains = self.balance - self.price
        self.balance = self.dividends_business = 0

    def reset(self):
        """
        Reset business value and capital gains to their initial values.
        """
        self.balance = self.init_balance
        self.cap_gains = self.init_balance - self.price



import pathlib
import setuptools

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setuptools.setup(
    name="cpr-rsi",
    version="1.2.1",
    description="Assessing Canadians' Preparation for Retirement",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://rsi-models.github.io/CPR/en",
    author="Team RSI",
    author_email="pyanni@gmail.com",
    license="MIT",
    classifiers=["Programming Language :: Python :: 3",
                 "License :: OSI Approved :: MIT License",
                 "Operating System :: OS Independent"],
    packages=setuptools.find_packages(),
    include_package_data=True, #probably useless
    install_requires=["pandas", "numpy", "requests", "statsmodels", "srpp",
                      "srd"],
    python_requires=">=3.6")


import csv


def get_params(file, numerical_key=False):
    """
    Function to create a dictionary from a .csv file.

    Parameters
    ----------
    file : _io.TextIOWrapper
        csv file
    numerical_key : bool, optional
        numerical key, by default False

    Returns
    -------
    dict
        Dictionary of parameters.
    """
    d_params = {}
    with open(file) as params:
        rows = csv.reader(params)
        next(rows)
        for row in rows:
            var, value, var_type, _ = row
            if var_type == 'int':
                d_params[var] = int(value)
            elif var_type == 'float':
                d_params[var] = float(value)
            elif var_type == 'bool':
                d_params[var] = bool(int(value))
            else:
                d_params[var] = value
    if numerical_key:
        d_params = {int(k): v for k, v in d_params.items()}
    return d_params


def add_params_as_attr(inst, file):
    """
    Function to add parameters to a class instance.

    Parameters
    ----------
    inst : object
        class instance
    file : _io.TextIOWrapper
        csv file
    """
    d_params = get_params(file)
    inst.__dict__.update(d_params)


def change_params(inst, extra_params):
    """
    Function to update the value of some parameters in inst with values from *extra_params*.

    Parameters
    ----------
    inst: dict
        parameters
    extra_params: dict
        parameters to be updated and their new values
    """
    new_params = {k: v for k, v in extra_params.items() if k in vars(inst)}
    for k, v in new_params.items():
        setattr(inst, k, v)


def create_nom_real(year, prices):
    """
    Creation of a function to convert prices from real to nominal and from nominal to real, with base year 2018.

    Parameters
    ----------
    year : int
        year
    prices : Prices
        instance of the class Prices

    Returns
    -------
    function
        Function converting nominal to real.
    function
        Function converting real to nominal.
    """
    factor = prices.d_infl_factors[year]

    def nom(x):
        return x * factor

    def real(x):
        return x / factor

    return nom, real


def create_nom(year, prices):
    """
    Creation of a function that converts prices from real to nominal, with base year 2018. Used for some specific purposes elsewhere in the code.

    Parameters
    ----------
    year : int
        year
    prices : Prices
        instance of the class Prices

    Returns
    -------
    function
        Function converting real to nominal.
    """
    factor = prices.d_infl_factors[year]

    def nom(x):
        return x * factor

    return nom


def create_real(year, prices):
    """
    Creation of a function that converts prices from nominal to real, with base year 2018.

    Parameters
    ----------
    year : int
        year
    prices : Prices
        instance of the class Prices

    Returns
    -------
    function
        function converting nominal to real
    """
    factor = prices.d_infl_factors[year]

    def real(x):
        return x / factor

    return real



import numpy as np
from CPR import assets
from CPR import debts


def create_hh(index, d_hh, common, prices):
    """
    Function that creates a household with one or two persons and attaches
    assets and debts. 

    Parameters
    ----------
    index : int
        household index
    d_hh : dict
        dictionary of household features
    common : Common
        instance of the class Common
    prices : Prices
        instance of the class Prices

    Returns
    -------
    Hhold
        Instance of the class Hhold.
    """
    # initialize household and spouses:
    l_hh = ['prov', 'couple', 'mix_bills', 'mix_bonds', 'mix_equity', 'fee',
            'fee_equity']
    l_sp = ['byear', 'sex', 'education', 'ret_age', 'claim_age_cpp',
            'init_wage', 'pension', 'init_room_rrsp', 'init_room_tfsa',
            'bal_rrsp', 'bal_tfsa', 'bal_other_reg', 'bal_unreg',
            'cap_gains_unreg', 'realized_losses_unreg', 'cont_rate_rrsp',
            'cont_rate_tfsa', 'cont_rate_other_reg', 'cont_rate_unreg',
            'withdrawal_rrsp', 'withdrawal_tfsa', 'withdrawal_other_reg',
            'withdrawal_unreg', 'replacement_rate_db', 'rate_employee_db',
            'income_previous_db', 'init_dc', 'rate_employee_dc',
            'rate_employer_dc']
    # create households with 1 or 2 people
    p = Person(d_hh, l_sp, common, prices)  # first p
    if not d_hh['couple']:
        hh = Hhold(d_hh, l_hh, index, common, p)
    else:  # p0 is first to retire
        l_s_p = ['s_' + var for var in l_sp]
        s_p = Person(d_hh, l_s_p, common, prices, s_=True)
        if p.ret_year <= s_p.ret_year:
            hh = Hhold(d_hh, l_hh, index, common, p, s_p)
        else:
            hh = Hhold(d_hh, l_hh, index, common, s_p, p)

    # initialize financial assets and rpp_dc:
    for p in hh.sp:
        p.fin_assets = {}
        if np.isnan(p.bal_tfsa):  # create tfsa account
            p.bal_tfsa, p.cont_rate_tfsa, p.withdrawal_tfsa = 0, 0, 0
        if np.isnan(p.bal_unreg):  # create unreg account
            p.bal_unreg, p.cont_rate_unreg, p.withdrawal_unreg = 0, 0, 0
            p.cap_gains_unreg, p.realized_losses_unreg = 0, 0

        p.fin_assets['unreg'] = assets.UnregAsset(p, hh, prices)
        for acc in ['rrsp', 'tfsa', 'other_reg']:
            if getattr(p, f'bal_{acc}') >= 0:
                p.fin_assets[acc] = assets.FinAsset(p, hh, acc)

    # initialize rpp_dc:
        if p.init_dc >= 0:
            p.rpp_dc = assets.RppDC(p, common)

    # initialize rpp_db:
        if (p.replacement_rate_db > 0) | (p.income_previous_db > 0):
            p.rpp_db = assets.RppDB(p)

    # initialize residences and business
    hh.residences = {}
    for resid in ['first_residence', 'second_residence']:
        if d_hh[resid] > 0:
            hh.residences[resid] = assets.RealAsset(d_hh, resid)
    if d_hh['business'] > 0:
        hh.business = assets.Business(d_hh)

    # initialize debts:
    hh.debts = {}
    l_debts = ['credit_card', 'personal_loan', 'student_loan', 'car_loan',
               'credit_line', 'first_mortgage', 'second_mortgage',
               'other_debt']
    for debt in l_debts:
        if d_hh[debt] > 0:
            hh.debts[debt] = debts.Debt(debt, d_hh, common, prices)
    return hh


class Person:
    """
    This class creates a person.

    Parameters
    ----------
    d_hh: dict
        dictionary containing all informations about households
    l_sp: list
        list of features attached to spouses
    common: Common
        instance of the class Common
    prices: Prices
        instance of the class Prices
    s_: bool
        False for head of household, True otherwise
    """
    def __init__(self, d_hh, l_sp, common, prices, s_=False):
        if s_:  # removes 's_' in front of variable names
            d_sp = {k[2:]: v for k, v in d_hh.items() if k in l_sp}
            self.who = 's_'  # to compare input and output data
        else:
            d_sp = {k: v for k, v in d_hh.items() if k in l_sp}
            self.who = ''
        self.__dict__.update(d_sp)
        self.byear = int(self.byear)  # s_byear is float because NaNs in column
        self.age = self.init_age = common.base_year - self.byear
        self.contribution_room = assets.ContributionRoom(self.init_room_rrsp,
                                                         self.init_room_tfsa)
        self.ret_age = min(int(self.ret_age), common.max_ret_age)
        self.ret_year = self.byear + self.ret_age
        self.retired = False
        # initialize cpp_qpp
        self.cpp = 0
        self.claim_age_cpp = np.clip(self.ret_age, common.min_claim_age_cpp,
                                     common.max_claim_age_cpp)
        # initialize annuities
        for acc in ['rrsp', 'rpp_dc', 'non_rrsp', 'non_rrsp_0']:
            setattr(self, f'annuity_{acc}_real', 0)
        # create wage_profile
        self.wage_profile = self.create_wage_profile(common, prices)

    def create_wage_profile(self, common, prices):
        """
        Function that creates wage profiles for each realisation of uncertainty, to be used when the stochastic version of the tool is selected.

        Parameters
        ----------
        common: Common
            instance of the class Common
        prices: Prices
            instance of the class Prices

        Returns
        -------
        np.array:
            Wage profiles.
        """
        rel_age = self.age - common.min_age_cpp
        rel_ret_age = self.ret_age - common.min_age_cpp
        rel_max_age = common.future_years - common.min_age_cpp

        log_wages = np.zeros(rel_max_age)
        diffs = prices.d_diff_log_wages[self.education]
        diffs = diffs[:rel_ret_age] - diffs[rel_age]
        log_wages[:rel_ret_age] = (np.log(self.init_wage) + diffs)
        log_wages = np.repeat(log_wages.reshape(-1, 1), common.nsim, axis=1)

        if common.non_stochastic is False:
            log_wages[rel_age+1:rel_ret_age, :] += \
                self.create_shocks(rel_ret_age-rel_age-1, common.nsim, prices)
            log_wages[:rel_age, :] += \
                np.flipud(self.create_shocks(rel_age, common.nsim, prices))

        return np.where(log_wages > 1, np.exp(log_wages), 0)

    def create_shocks(self, T, N, prices):
        """
        Function that creats a time series of shocks to the wage.

        Parameters
        ----------
        T: int
            length of the time series
        N: int
            number of realisations
        prices: Prices
            instance of the class Prices

        Returns
        -------
        np.array:
            Time series of shocks to the wage.
        """
        if T == 0:
            return np.zeros((1, N))
        eps_trans = np.random.normal(0, prices.sig_trans_wage, (T, N))
        eps_pers = np.random.normal(0, prices.sig_pers_wage, (T, N))
        shock_pers = np.zeros((T, N))
        shock_pers[0, :] = eps_pers[0, :]
        for t in range(1, T):
            shock_pers[t, :] = prices.rho_wage * shock_pers[t-1, :] \
                + eps_pers[t, :]
        return eps_trans + shock_pers


class Hhold():
    """
    This class creates a household.

    Parameters
    ----------
    d_hh: dict
        dictionary containing all informations about households
    l_hhold: list
        list of features attached to the household
    index: int
        household index
    common: Common
        instance of the class Common
    p0: Person
        first spouse
    p1: Person
        second spouse (in the case of a couple)
    """
    def __init__(self, d_hh, l_hhold, index, common, p0, p1=None):
        self.sp = [p0] if p1 is None else [p0, p1]
        d_hhold = {k: v for k, v in d_hh.items() if k in l_hhold}
        self.__dict__.update(d_hhold)
        self.index = index
        self.set_other_years(common)

    def set_other_years(self, common):
        """
        Function that sets years of partial and full retirement as well as years in which pre-retirement and post-retirement consumptions are assessed.

        Parameters
        ----------
        common: Common
            instance of the class Common
        """
        # consumption before retirement at age_cons_bef_ret but after 2018
        # and before first retirement
        self.cons_bef_ret_year = np.clip(
            self.sp[0].byear + common.age_cons_bef_ret,
            common.base_year, self.sp[0].ret_year - 1)

        if self.couple == False:
            self.ret_year = self.sp[0].ret_year
            self.cons_after_ret_year = max(
                self.sp[0].byear + common.official_ret_age, self.ret_year)
        else:
            self.ret_year = self.sp[1].ret_year  # p[1] is last to retire
            if self.sp[0].ret_year < self.sp[1].ret_year:
                self.partial_ret_year = self.sp[0].ret_year
            self.cons_after_ret_year = max(
                self.sp[1].byear + common.official_ret_age, self.ret_year)
